/*!
* \file PropagationNMPB08.cpp
* \brief Definition of the library functions that can be called by external software
* \author CSTB
* \version 1.0
*/

/*
 * in order to use specific memory allocators (DL, NED or SYSTEM):
 *	- compile with options -D CPP_NEW_DELETE=<XX>_MEMORY
 *  - link with <XX>Memory.DLL
 */
#include "PropagationNMPB08.h"
#include "CalculPropagation.h"
#include "SousCalculs/ElementaryPath.h"
#include <vector>
#include <math.h>
#include <stdio.h>

//#define _TRACE_MEMORY_
//#define CPP_NEW_DELETE SYSTEM_MEMORY

#ifdef _TRACE_MEMORY_

int		nb_alloc = 0 ;
size_t	size_alloc = 0 ;
int		nb_free = 0 ;
size_t	size_free = 0 ;

static size_t bloc_count [1024] ;

void* operator new (size_t n)
{
	nb_alloc++ ;
	size_alloc += n ;
	bloc_count[n]++ ;
	void* b = malloc (n) ;
	memset (b,0,n) ;
	return b ;
}

void operator delete (void* b)
{
	if (b == 0) return ;
	size_t n = _msize(b) ; 
	nb_free++ ;
	size_free += n ;
	bloc_count[n]-- ;
	free(b) ;
}

void* operator new[] (size_t n)
{
	nb_alloc++ ;
	size_alloc += n ;
	bloc_count[n]++ ;
	void* b = malloc (n) ;
	memset (b,0,n) ;
	return b ;
}

void operator delete[] (void* b)
{
	if (b == 0) return ;
	size_t n = _msize(b) ; 
	nb_free++ ;
	size_free += n ;
	bloc_count[n]-- ;
	free(b) ;
}

void show_mem (const char* where)
{
	printf ("%s\n", where) ;
	printf (". alloc = %8d bytes in %5d blocs \n", size_alloc, nb_alloc) ;
	printf (". free  = %8d bytes in %5d blocs \n", size_free,  nb_free) ;
	printf (". using = %8d bytes in %5d blocs \n", size_alloc - size_free, nb_alloc - nb_free) ;
	for (int i = 0 ; i < 1024 ; ++i)
	{
		if (bloc_count[i] == 0) continue ;
		printf (". using %d blocs of size %d bytes \n", bloc_count[i],i) ;
	}
	printf ("\n") ;
}

#else

static void show_mem (const char* where)
{
}

#ifdef CPP_NEW_DELETE
#include "SystemMemory.h"
#endif

#endif

using namespace std;
using namespace CalculPropagationNMPB;
using namespace ElementaryPathNMPB;

/*!
 * \brief Check if the path is OK
 * 
 * \param path the path
 * 
 * \returns true if all is OK
 */
bool CheckPath(PathID path)
{
	if(path == NULE)
	{
		printf ("ERROR : Chemin non initialise \n") ;
		return false;
	}

	return true;
}

/*!
 * \brief Creates path calculator with default frequency range
 * 
 * \returns the new Propagation Path
 */
PathID NMPB08_CreatePath()
{
	show_mem ("CreatePath") ;

	try
	{
		PropagationPath* path = new PropagationPath;

		// Default frequencies : 18 nominal median frequencies :
		double NominalMedianFrequency[] = {100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000};
		const int frequencyNumber = 18;
		path->SetFrequencies(frequencyNumber, NominalMedianFrequency) ;
		/*
		 * modification 25/03/2013
		 * 
		 * A la demande du LRS, la correction de faible hauteur est DESACTIVE par dfaut !
		 */
		path->SetOption (FORCE_CH_EQUAL_ONE, true) ;
		return path;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return NULE;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return NULE;
	}
}

/*!
 * \brief Creates path calculator with user defined frequency range
 * 
 * \param nbFreq : frequencies number
 * 
 * \param freq : frequencies values
 * 
 * \returns the new Propagation Path
 */
PathID NMPB08_CreatePathEx(int nbFreq, double const* freq)
{
	show_mem ("CreatePathEx") ;

	try
	{
		PropagationPath* path = new PropagationPath();
		path->SetFrequencies(nbFreq, freq);
		/*
		 * modification 25/03/2013
		 * 
		 * A la demande du LRS, la correction de faible hauteur est DESACTIVE par dfaut !
		 */
		path->SetOption (FORCE_CH_EQUAL_ONE, true) ;

		return path;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return NULE;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return NULE;
	}
}

/*!
 * \brief Set an option for the path
 * 
 * \param path The path to set the option
 * 
 * \param option The option to set
 * 
 * \param on_off True if the option must be set to true
 * 
 * \returns true if all OK
 */
bool NMPB08_SetOption(PathID path, Option option, bool on_off)
{
	try
	{
		if(!CheckPath(path)) return false;

		PropagationPath* propagationPath = (PropagationPath*) path;
		propagationPath->SetOption(option, on_off);
		return true;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return false;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return false;
	}
}

/*!
 * \brief Gets the frequencies number
 * 
 * \param path The path containing frequencies
 * 
 * \returns the frequencies number
 */
int NMPB08_GetNbFrequencies (PathID path)
{
	try
	{
		if(!CheckPath( path)) return 0;
	
		PropagationPath* propagationPath = (PropagationPath*) path;
		return propagationPath->GetNbFrequencies();
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
* \brief Gets the frequencies array
* 
* \param path The path containing frequencies
* 
* \returns the frequencies array
*/
double const* NMPB08_GetFrequencies (PathID path)
{
	try
	{
		if(!CheckPath( path)) return 0;

		PropagationPath* propagationPath = (PropagationPath*) path;
		return propagationPath->GetFrequencies();
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
 * \brief Delete the path calculator
 * 
 * \param path The path to delete
 * 
 * \returns true if all OK
 */
bool NMPB08_DeletePath (PathID path)
{
	try
	{
		if(!CheckPath( path)) return false;

		PropagationPath* propagationPath = (PropagationPath*) path;
		delete propagationPath;
		
		show_mem ("DeletePath") ;

		return true;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return false;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return false;
	}
}

/*!
* \brief clear the path profile
* 
* \param path The path to clear
* 
* \returns true if all OK
*/
bool NMPB08_ClearPath (PathID path)
{
	show_mem ("ClearPath (start)") ;
	try
	{	
		if(!CheckPath( path)) return false;

		PropagationPath* propagationPath = (PropagationPath*) path;
		propagationPath->ClearPath();
		show_mem ("ClearPath (end)") ;

		return true;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return false;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return false;
	}
}

/*!
 * \brief Add a segment to the path profile
 * 
 * \param path The path to add segment
 * 
 * \param point3D The 3D coordinates to add
 * 
 * \param g The impedance value for the segment before the added point
 * 
 * \returns true if all OK
 */
bool NMPB08_ExtendPath (PathID path, Position3D const* point3D, double g)
{
	try
	{
		if(!CheckPath( path)) return false;

		PropagationPath* thePath = (PropagationPath*) path;
		thePath->ExtendPath(point3D, g);
		return true;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return false;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return false;
	}
}

/*!
 * \brief Add a segment to the path profile with extension data
 * 
 * \param path The path to add segment.
 * 
 * \param point3D The 3D coordinates to add
 * 
 * \param g The impedance value for the segment before the added point
 * 
 * \param ext The extension data
 * 
 * \returns true if all OK
 */
bool NMPB08_ExtendPathExt (PathID path, Position3D const* point3D, double g, ExtensionNMPB const* ext) 
{
	try
	{
		if(!CheckPath( path)) return false;

		PropagationPath* thePath = (PropagationPath*) path;
		thePath->ExtendPathExt(point3D, g, ext);
		return true;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return false;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return false;
	}
}

/*!
 * \brief Set the source height
 * 
 * \param path The path to set source height
 * 
 * \param h The source height
 * 
 * \returns the exception number (0 if no exception)
 */
int NMPB08_SetSourceHeight (PathID path, double h)
{
	try
	{
		if(!CheckPath( path)) return ERRNullPath;

		PropagationPath* thePath = (PropagationPath*) path;
		thePath->SetSourceHeight(h);
		return ERRNone;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return *err;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return ERRUnknown;
	}
}

/*!
* \brief Set the receiver height
* 
* \param path The path to set receiver height
* 
* \param h The receiver height
* 
* \returns the exception number (0 if no exception)
*/
int NMPB08_SetReceiverHeight (PathID path, double h)
{
	try
	{
		if(!CheckPath( path)) return ERRNullPath;

		PropagationPath* thePath = (PropagationPath*) path;
		thePath->SetReceiverHeight(h);
		return ERRNone;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return *err;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return false;
	}
}
/*
double MeanAttenuation (double const* attenuations, int nbAtt)
{
	try
	{
		if (nbAtt <= 0)
		{
			return 0;
		}

		double meanAtt = 0;
		for (int i = 0 ; i < nbAtt ; i++)
		{
			meanAtt += attenuations[i];
		}

		return (meanAtt / nbAtt);
	}
	catch ( const int * )
	{
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}
*/

/*!
 * \brief Do the propagation calculation
 * 
 * \param path The path to do propagation calculation
 * 
 * \returns the exception number (0 if no exception)
 */
int NMPB08_DoCalculation (PathID path)
{
	show_mem ("DoCalculation (start)") ;
	try
	{
		if(!CheckPath( path)) return ERRNullPath;

		PropagationPath* propagationPath = (PropagationPath*) path;
		//propagationPath->favorableAttenuations = 0 ;
		//propagationPath->homogeneousAttenuations = 0 ;
		
		double* favorableAttenuations   = propagationPath->favorableAttenuations;
		double* homogeneousAttenuations = propagationPath->homogeneousAttenuations ;

		// Calculate elementary paths between source and receiver :
		SetElementaryPath(propagationPath);
		if(propagationPath->GetOption(TRACE_DETAILS))
		{
			printf("Distance SR : %7.3f \n", propagationPath->distSR);
		}

		const int frequencyNumber = (int) propagationPath->frequencies.size();
		//double* favorableAttenuations = new double[frequencyNumber];
		//double* homogeneousAttenuations = new double[frequencyNumber];

		double* divergenceAttenuations = new double[frequencyNumber];
		double* atmosphericAbsorptions = new double[frequencyNumber];
		double* boundaryAttenuations_h = new double[frequencyNumber];
		double* boundaryAttenuations_f = new double[frequencyNumber];
		double* groundAttenuations_h = new double[frequencyNumber];
		double* groundAttenuations_f = new double[frequencyNumber];
		double* diffractionAttenuations_h = new double[frequencyNumber];
		double* diffractionAttenuations_f = new double[frequencyNumber];

		for(int i = 0 ; i < frequencyNumber ; i++)
		{
			//total attenuation along the propagation path
			Attenuation pathAttenuation(propagationPath, propagationPath->frequencies[i]);
			pathAttenuation.AttenuationCalculation();
			//Ai,F
			favorableAttenuations[i] = pathAttenuation.getFavorableAttenuation();
			//Ai,H
			homogeneousAttenuations[i] = pathAttenuation.getHomogeneousAttenuation();

			// used values :
			divergenceAttenuations[i] = pathAttenuation.getDivergenceAttenuation();
			atmosphericAbsorptions[i] = pathAttenuation.getAtmosphericAbsorption();
			boundaryAttenuations_h[i] = pathAttenuation.getBoundaryAttenuation_h();
			boundaryAttenuations_f[i] = pathAttenuation.getBoundaryAttenuation_f();
			groundAttenuations_h[i] = pathAttenuation.getGroundAttenuation_h() + pathAttenuation.getEmbankmentAttenuation();
			groundAttenuations_f[i] = pathAttenuation.getGroundAttenuation_f() + pathAttenuation.getEmbankmentAttenuation();
			diffractionAttenuations_h[i] = pathAttenuation.getDiffractionAttenuation_h();
			diffractionAttenuations_f[i] = pathAttenuation.getDiffractionAttenuation_f();
		}

		//propagationPath->favorableAttenuations = favorableAttenuations;
		//propagationPath->homogeneousAttenuations = homogeneousAttenuations;

		// Sum on frequencies :
		/*
		double Adiv_sum = MeanAttenuation(divergenceAttenuations, frequencyNumber);
		double Aatm_sum = MeanAttenuation(atmosphericAbsorptions, frequencyNumber);
		double Asol_h_sum = MeanAttenuation(groundAttenuations_h, frequencyNumber);
		double Adif_h_sum = MeanAttenuation(diffractionAttenuations_h, frequencyNumber);
		double Att_h_sum = MeanAttenuation(homogeneousAttenuations, frequencyNumber);
		double Asol_f_sum = MeanAttenuation(groundAttenuations_f, frequencyNumber);
		double Adif_f_sum = MeanAttenuation(diffractionAttenuations_f, frequencyNumber);
		double Att_f_sum = MeanAttenuation(favorableAttenuations, frequencyNumber);
		*/

		// write data :
		if(propagationPath->GetOption(TRACE_DETAILS))
		{
			printf("Conditions homogenes : \n");
			printf("      f;   Adiv;   Aatm;   Asol;   Adif;     Ah\n");
			//printf("      A;%7.1f;%7.1f;%7.1f;%7.1f;%7.1f \n", Adiv_sum, Aatm_sum, Asol_h_sum, Adif_h_sum, Att_h_sum);
			for (int i = 0 ; i < frequencyNumber ; i++)
			{
				printf ("%7.0f;%7.1f;%7.1f;", propagationPath->frequencies[i], divergenceAttenuations[i], atmosphericAbsorptions[i]) ;
				printf ("%7.1f;%7.1f;%7.1f", groundAttenuations_h[i], diffractionAttenuations_h[i], homogeneousAttenuations[i]) ;
				printf(" \n");
			}
			printf(" \n");
			printf("Conditions favorables : \n");
			printf("      f;   Adiv;   Aatm;   Asol;   Adif;     Af\n");
			// printf("      A;%7.1f;%7.1f;%7.1f;%7.1f;%7.1f \n", Adiv_sum, Aatm_sum, Asol_f_sum, Adif_f_sum, Att_f_sum);
			for (int i = 0 ; i < frequencyNumber ; i++)
			{
				printf ("%7.0f;%7.1f;%7.1f;", propagationPath->frequencies[i], divergenceAttenuations[i], atmosphericAbsorptions[i]) ;
				printf ("%7.1f;%7.1f;%7.1f", groundAttenuations_f[i], diffractionAttenuations_f[i], favorableAttenuations[i]) ;
				printf(" \n");
			}
			printf(" \n");
		}


		delete [] divergenceAttenuations;
		delete [] atmosphericAbsorptions;
		delete [] boundaryAttenuations_h;
		delete [] boundaryAttenuations_f;
		delete [] groundAttenuations_h;
		delete [] groundAttenuations_f;
		delete [] diffractionAttenuations_h;
		delete [] diffractionAttenuations_f;

		show_mem ("DoCalculation (end)") ;

		return ERRNone;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return *err;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return ERRUnknown;
	}
}	

/*!
* \brief Get path attenuations under favorable conditions
*
* \param path The path to get attenuations
* 
* \returns the favorable attenuation for each frequency range
*/
double const* NMPB08_GetAttF (PathID path)
{
	try
	{
		if(!CheckPath( path)) return 0;

		PropagationPath* propagPath = (PropagationPath*) path;
		return propagPath->favorableAttenuations;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
* \brief Get path attenuations under homogeneous conditions
*
* \param path The path to get attenuations
* 
* \returns the homogeneous attenuation for each frequency range
*/
double const* NMPB08_GetAttH (PathID path)
{
	try
	{
		if(!CheckPath( path)) return 0;

		PropagationPath* propagPath = (PropagationPath*) path;
		return propagPath->homogeneousAttenuations;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
 * \brief Calculates long-term sound level for the given the sound levels (homogeneous and favorable)
 * 
 * \param soundLevel_h : sound level in homogeneous conditions
 * 
 * \param soundLevel_f : sound level in downward-refraction conditions
 * 
 * \param p The average occurrence of downward-refraction conditions in the direction of the path
 * 
 * \returns the calculated sound level
 */
double NMPB08_Leq_LT(double soundLevel_h, double soundLevel_f, double p)
{
	try
	{
		return SoundLevelForPath(soundLevel_h, soundLevel_f, p);
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
* \brief Calculates sound levels sum
*
* Used for  5.2.4 Formula (8) and  5.2.5 Formula (9)
* 
* \param n the sound levels number
* 
* \param levels The sound levels to sum
* 
* \returns the sum
*/
double NMPB08_SumLevels(int n, double const* levels)
{
	try
	{
		return SumLevels(n, levels);
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
* \brief Calculates the favorable conditions probability for the (SR) direction
* 
* \param source The source position 3D
* 
* \param receiver The receiver position 3D
* 
* \param nbAngles The number of angle probabilities (should be 18)
* 
* \param fcpAngles The favorable conditions probabilities for the angles 20, 40, 60, ..., 360
* 
* \param angleNorth The north direction (Ox,ON)
* 
* \returns the favorable conditions probability for the (SR) direction
*/
double NMPB08_GetFavorableConditionProbability(Position3D const* source, Position3D const* receiver, int nbAngles, const double* fcpAngles, double angleNorth)
{
	try
	{
		if (nbAngles != 18)
		{
			printf (".ERROR: NMPB requires fpc tables in steps of 20 \n") ;
			return 0;
		}

		return GetFavorableConditionProbability(source, receiver, nbAngles, fcpAngles, angleNorth);
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return 0;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return 0;
	}
}

/*!
* \brief Calculates the Long-term sound level Leq due to one source at point R, in each given frequency band.
* 
* \param nbFreq The frequencies number (user data)
* 
* \param Lw The sound power levels of the source, in each frequency band (user data)
* 
* \param attH Attenuations due to the propagation between source and receiver in homogeneous conditions, in each frequency band (user data)
* 
* \param attF Attenuations due to the propagation between source and receiver in downward-refraction conditions, in each frequency band (user data)
* 
* \param fcp Probability of occurrence of downward-refraction conditions over a long-term period in a given direction, p in [0, 1] (user data)
* 
* \param LeqH Sound levels due to source Si at point R in homogeneous conditions, in each frequency band (calculated in this function)

* 
* \param LeqF Sound levels due to source Si at point R in downward-refraction conditions, in each frequency band (calculated in this function)
* 
* \param LeqLT Long-term sound levels due to source Si at point R, in each given frequency band (calculated in this function)
* 
* \returns the exception number (0 if no exception)
*/
int NMPB08_CalculateLeqLT(int nbFreq, double const* Lw, double const* attH, double const* attF, double fcp, double* LeqH, double* LeqF, double* LeqLT)
{
	try
	{
		CalculateLeqLT(nbFreq, Lw, attH, attF, fcp, LeqH, LeqF, LeqLT);
		return ERRNone;
	}
	catch ( const int * err)
	{
		printf ("NMPB08_Propagation returned error = %d", *err) ;
		return *err;
	}
	catch (const std::exception& e)
	{
		printf ("ERROR : %s \n", e.what()) ;
		return ERRUnknown;
	}
}
