#ifndef _NMPB08_PATHDEF_INCLUDED_
#define _NMPB08_PATHDEF_INCLUDED_

/*!
* \file PropagationNMPB08.h
* \brief Definition of the library functions that can be called by external software
* \author CSTB
* \version 1.0
*
* \mainpage NMPB08 Propagation Model (as published in NF S 31-133:2011)
	\section main Public functions defined in the PropagationNMPB08 library
	\subsection sub1 Create and delete a calculation context
	- \ref NMPB08_CreatePath
	- \ref NMPB08_DeletePath
	- \ref NMPB08_CreatePathEx
	- \ref NMPB08_SetOption
	- \ref NMPB08_GetNbFrequencies
	- \ref NMPB08_GetFrequencies
	\subsection sub2 Define the propagation path
	- \ref NMPB08_ClearPath
	- \ref NMPB08_ExtendPath
	- \ref NMPB08_ExtendPathExt
	- \ref NMPB08_SetSourceHeight
	- \ref NMPB08_SetReceiverHeight
	\subsection sub3 Calculate propagation attenuation
	- \ref NMPB08_DoCalculation
	- \ref NMPB08_GetAttH
	- \ref NMPB08_GetAttF
	\subsection sub4 Calculate long term averaged noise levels
	- \ref NMPB08_GetFavorableConditionProbability 
	- \ref NMPB08_CalculateLeqLT 
	- \ref NMPB08_SumLevels
	- \ref NMPB08_Leq_LT
*/

#include "pathdefNMPB.h"

#ifndef _COMPILE_NMPB
	#ifdef _WIN32 
		#define _COMPILE_NMPB extern "C" __declspec(dllexport)
	#else
		#define _COMPILE_NMPB extern "C"
	#endif 

#endif

/*!
 * \brief the path reference
 */
typedef void* PathID ;

/*!
* \brief Creates path calculator with default frequency range
* 
* \returns the new Propagation Path
*/
_COMPILE_NMPB PathID NMPB08_CreatePath () ;

/*!
* \brief Creates path calculator with user defined frequency range
* 
* \param nbFreq : frequencies number
* 
* \param freq : frequencies values
* 
* \returns the new Propagation Path
*/
_COMPILE_NMPB PathID NMPB08_CreatePathEx (int nbFreq, double const* freq) ;

/*!
* \brief Set an option for the path
* 
* \param path The path to set the option
* 
* \param option The option to set
* 
* \param on_off True if the option must be set to true
* 
* \returns true if all OK
*/
_COMPILE_NMPB bool NMPB08_SetOption(PathID, Option option, bool on_off);

/*!
* \brief Gets the frequencies number
* 
* \param path The path containing frequencies
* 
* \returns the frequencies number
*/
_COMPILE_NMPB int NMPB08_GetNbFrequencies (PathID);

/*!
* \brief Gets the frequencies array
* 
* \param path The path containing frequencies
* 
* \returns the frequencies array
*/
_COMPILE_NMPB double const* NMPB08_GetFrequencies (PathID);

/*!
* \brief Delete the path calculator
* 
* \param path The path to delete
* 
* \returns true if all OK
*/
_COMPILE_NMPB bool	NMPB08_DeletePath (PathID) ;

/*!
* \brief clear the path profile
* 
* \param path The path to clear
* 
* \returns true if all OK
*/
_COMPILE_NMPB bool 	NMPB08_ClearPath (PathID) ;

/*!
* \brief Add a segment to the path profile
* 
* \param path The path to add segment
* 
* \param point3D The 3D coordinates to add
* 
* \param g The impedance value for the segment before the added point
* 
* \returns true if all OK
*/
_COMPILE_NMPB bool	NMPB08_ExtendPath (PathID, Position3D const* point3D, double g) ;

/*!
* \brief Add a segment to the path profile with extension data
* 
* \param path The path to add segment.
* 
* \param point3D The 3D coordinates to add
* 
* \param g The impedance value for the segment before the added point
* 
* \param ext The extension data
* 
* \returns true if all OK
*/
_COMPILE_NMPB bool	NMPB08_ExtendPathExt (PathID, Position3D const* point3D, double g, ExtensionNMPB const* ext) ;

/*!
* \brief Set the source height
* 
* \param path The path to set source height
* 
* \param h The source height
* 
* \returns the exception number (0 if no exception)
*/
_COMPILE_NMPB int	NMPB08_SetSourceHeight (PathID, double h) ;

/*!
* \brief Set the receiver height
* 
* \param path The path to set receiver height
* 
* \param h The receiver height
* 
* \returns the exception number (0 if no exception)
*/
_COMPILE_NMPB int	NMPB08_SetReceiverHeight (PathID, double h) ;

/*!
* \brief Do the propagation calculation
* 
* \param path The path to do propagation calculation
* 
* \returns the exception number (0 if no exception)
*/
_COMPILE_NMPB int	NMPB08_DoCalculation (PathID) ;

/*!
* \brief Get path attenuations under favorable conditions
*
* \param path The path to get attenuations
* 
* \returns the favorable attenuation for each frequency range
*/
_COMPILE_NMPB double const* NMPB08_GetAttF (PathID) ;

/*!
* \brief Get path attenuations under homogeneous conditions
*
* \param path The path to get attenuations
* 
* \returns the homogeneous attenuation for each frequency range
*/
_COMPILE_NMPB double const* NMPB08_GetAttH (PathID) ;

/*!
* \brief Calculates long-term sound level for the given the sound levels (homogeneous and favorable)
* 
* \param soundLevel_h : sound level in homogeneous conditions
* 
* \param soundLevel_f : sound level in downward-refraction conditions
* 
* \param p The average occurrence of downward-refraction conditions in the direction of the path
* 
* \returns the calculated sound level
*/
_COMPILE_NMPB double NMPB08_Leq_LT(double soundLevel_h, double soundLevel_f, double p) ;

/*!
* \brief Calculates sound levels sum
*
* Used for  5.2.4 Formula (8) and  5.2.5 Formula (9)
* 
* \param n the sound levels number
* 
* \param levels The sound levels to sum
* 
* \returns the sum
*/
_COMPILE_NMPB double NMPB08_SumLevels(int n, double const* levels);

/*!
* \brief Calculates the favorable conditions probability for the (SR) direction
* 
* \param source The source position 3D
* 
* \param receiver The receiver position 3D
* 
* \param nbAngles The number of angle probabilities (should be 18)
* 
* \param fcpAngles The favorable conditions probabilities for the angles 20, 40, 60, ..., 360
* 
* \param angleNorth The north direction (Ox,ON)
* 
* \returns the favorable conditions probability for the (SR) direction
*/
_COMPILE_NMPB double NMPB08_GetFavorableConditionProbability (Position3D const* source, Position3D const* receiver, int nbAngles, double const* fcpAngles, double angleNorth);

/*!
* \brief Calculates the Long-term sound level Leq due to one source at point R, in each given frequency band.
* 
* \param nbFreq The frequencies number (user data)
* 
* \param Lw The sound power levels of the source, in each frequency band (user data)
* 
* \param attH Attenuations due to the propagation between source and receiver in homogeneous conditions, in each frequency band (user data)
* 
* \param attF Attenuations due to the propagation between source and receiver in downward-refraction conditions, in each frequency band (user data)
* 
* \param fcp Probability of occurrence of downward-refraction conditions over a long-term period in a given direction, p in [0, 1] (user data)
* 
* \param LeqH Sound levels due to source Si at point R in homogeneous conditions, in each frequency band (calculated in this function)
* 
* \param LeqF Sound levels due to source Si at point R in downward-refraction conditions, in each frequency band (calculated in this function)
* 
* \param LeqLT Long-term sound levels due to source Si at point R, in each given frequency band (calculated in this function)
* 
* \returns the exception number (0 if no exception)
*/
_COMPILE_NMPB int NMPB08_CalculateLeqLT (int nbFreq, double const* Lw, double const* attH, double const* attF, double fcp, double* LeqH, double* LeqF, double* LeqLT);



#endif

