/*!
* \file ElementaryPath.cpp
* \brief Elementary path determination (3D -> 2D, convex hull)
* \author CSTB
* \version 1.0
*/

#include "ElementaryPath.h"

//#define _CHECK_FLATNESS_NMPB96_
//#define _CHECK_FLATNESS_CSTB_

#include "math.h"

// angle_north calcule l'angle en degrs entre l'axe Y et le vecteur (x,y) 
// l'angle est mesur dans le sens des aiguilles d'une montre
// l'angle est positif lorsque x > 0 et ngatif lorsque x < 0

static double angle_north (double x, double y)
{
	return atan2 (x,y) * 180/PI ;
}

static const double angle_ext_min = 181 ;

namespace ElementaryPathNMPB
{
	/*!
	 * \brief Finds the convex hull of the screenItems vector
	 * 
	 * p.47-49 -  9.4.3
	 *
	 * \param pathItems The points to find the convex hull
	 * 
	 * \param n1 The first element place in the vector
	 * 
	 * \param n2 The last element place in the vector
	 * 
	 * \param level The convex hull level
	 */
	void ConvexHull(vector<ProfilePointNMPB*>& pathItems, int n1, int n2, int level)
	{
		int      i;
		double   d1,z1,d2,z2;
		int      initS;
		double   dist, zray, zmax, zdif;

		// ray equation & height for the convex hull
		d1 = pathItems[n1]->position2D.d ;
		z1 = pathItems[n1]->position2D_withHeight().z ;

		d2 = pathItems[n2]->position2D.d ;
		z2 = pathItems[n2]->position2D_withHeight().z ;

		for (i = n1+1 ; i < n2 ; i++)
		{
		  dist = pathItems[i]->position2D.d ;
		  zray = z1 + (z2 - z1) * (dist - d1) / (d2 - d1) ;
		  pathItems[i]->h_ray = zray - pathItems[i]->position2D_withHeight().z ;
		}

		// seek the higher point to the ray
		initS = 0  ;
		zmax = 0. ;

		for (i = n1+1 ; i < n2 ; i++)
		{
			// if point is not reflection or side diffraction
			if (pathItems[i]->ext == NULL || 
				(pathItems[i]->ext->type != ETReflection_NMPB && pathItems[i]->ext->type != ETSideDiffraction_NMPB))
			{
				zdif = -pathItems[i]->h_ray ;

				if (initS == 0 || zdif > zmax)
				{
					initS = i;
					zmax = zdif;
				}
			}
		}

		// if the higher point is above the ray : it is kept as diffracted 
		//	and the function is called on the left and on the right of this point
		if (zmax > 0)
		{
		  pathItems[initS]->isDiff = 1 ;

		  ConvexHull(pathItems ,n1 , initS, level+1);
		  ConvexHull(pathItems ,initS , n2, level+1);

		  return ;
		}

		// if the higher point is under the SR ray (level 0) but not too low : it is kept as diffracted
		if (level == 0 && initS != 0)
		{
#ifdef _CHECK_FLATNESS_NMPB96_
		/*
		 * critre propos par Yannick Gabillet pour Mithra V2.0
		 */
			zdif = - pathItems[initS]->h_ray ;
			zdif = zdif / sqrt ( 0.34 * (pathItems[initS]->position2D->d - d1) * (d2 - pathItems[initS]->position2D->d) / (d2 - d1)) ;
			if (zdif > -0.5)  pathItems[initS]->isDiff = 1 ;
#else

#ifdef _CHECK_FLATNESS_CSTB_
			/*
			 * critre propos par	Dirk Van Maercke (voir change mail 30/01/2013 - 06/02/2013)
			 *
			 * mme test que ci-dessus, mais en prenant comme rfrence la projection au sol
			 * de la source et du rcepteur (donc le terrain dans son ensemble)
			 */
			int i1 = n1 ;   
			int i2 = n2 ;   
#else
			/*
			 * critre propos par Guillaume Dutillieux (voir change mail 30/01/2013 - 06/02/2013)
			 */
			int i1 = initS-1 ;   
			int i2 = initS+1 ;   
#endif
			double d1 = pathItems[i1]->position2D.d ;
			double z1 = pathItems[i1]->position2D.z ;
			double d2 = pathItems[i2]->position2D.d ;
			double z2 = pathItems[i2]->position2D.z ;
			double dp = pathItems[initS]->position2D.d ;
			double zp = pathItems[initS]->position2D.z ;
			
			double a1 = angle_north (dp-d1, z1-zp) ; // angle  gauche relatif  la verticale au point P[k]
			double a2 = angle_north (d2-dp, z2-zp) ; // angle  droite relatif  la verticale au point P[k]
			
			//printf ("a1 = %.1f a2 = %.1f \n", a1, a2) ;
			if (a2 + a1 > angle_ext_min) pathItems[initS]->isDiff = 1 ;
			
			//printf ("a1+a2 = %.1f => dif below line of sight : %s \n", a1+a2, pathItems[initS]->isDiff ? "YES" : "NO") ;
#endif
		}
	}

	/*!
	 * \brief Sets the elementary path for the given path (calculates convex hull and plane positions)
	 * 
	 * \param path The Propagation path
	 * 
	 * \throws ERRNoPoint
	 * 
	 * \throws ERROnePoint
	 */
	void SetElementaryPath(PropagationPath* path)
	{
		// check there are at least 2 elements in the path :
		int pointsNumber = (int) path->pathPoints.size();
		if (pointsNumber < 1)
		{
			printf (".ERROR: path empty \n") ;
			throw new int( ERRNoPoint );
		}
		else if (pointsNumber < 2)
		{
			printf (".ERROR: only one point in the path \n") ;
			throw new int( ERROnePoint );
		}

		// 3D -> 2.5D (all points will be in a plane ; if vertical reflection or lateral diffraction, image points are calculated and distances are cumulated) :
		ProfilePointNMPB* source = path->pathPoints[0];
		source->position2D.d = 0;
		source->position2D.z = source->position3D.z;
		source->isDiff = 0;
		source->h_ray = 0;

		int i_precReflexion = 0; // source is considered as first reflection
		double accuDistance = 0; // accumulated ground distance
		for (int i = 1 ; i < pointsNumber ; i++)
		{
			// if point is reflection, side diffraction, or is receiver :
			if ((path->pathPoints[i]->ext && path->pathPoints[i]->ext->type == ETReflection_NMPB) || 
				(path->pathPoints[i]->ext && path->pathPoints[i]->ext->type == ETSideDiffraction_NMPB) || 
				i == pointsNumber - 1)
			{
				// determine 2D position for all the points between the 2 reflections :
				ProfilePointNMPB* prec_screen = path->pathPoints[i_precReflexion];
				ProfilePointNMPB* cur_screen = path->pathPoints[i];
				
				for (int j = i_precReflexion + 1 ; j < i ; j++)
				{
					ProfilePointNMPB* pathPoint = path->pathPoints[j];
					FillPlanePosition(pathPoint, prec_screen, accuDistance);
					pathPoint->isDiff = 0;
					pathPoint->h_ray = 0;
				}

				FillPlanePosition(cur_screen, prec_screen, accuDistance);
				cur_screen->isDiff = 0;
				cur_screen->h_ray = 0;

				accuDistance = cur_screen->position2D.d;
				i_precReflexion = i;			
			}
		}

		// convex hull for vertical diffractions
		ConvexHull(path->pathPoints, 0, pointsNumber - 1, 0);

		// SR distance (|S'R'| if reflections) : SR distance with the enlarged trajectory taking account of reflections and lateral diffractions (but not vertical diffractions)
		Position2D p1 = source->position2D_withHeight() ;
		Position2D p2 = path->pathPoints[pointsNumber-1]->position2D_withHeight();
		path->distSR = distance2D(&p1, &p2) ;
	}
}