/*!
* \file Embankment.cpp
* \brief Calculation embankment attenuation
* \author CSTB
* \version 1.0
*/

#include "Embankment.h"
#include <math.h>
#include <stdio.h>

namespace EmbankmentNMPB
{
	/*!
	* \brief Initialization of the Embankment class and embankment attenuation calculation
	* 
	* \param terrainItems the terrain elements
	* 
	* \param freq the frequency
	* 
	* \throws ERRFrequency
	*/
	Embankment::Embankment(vector<ProfilePointNMPB*> terrainItems, double freq)
	{
		_embankmentAttenuation = 0;
		_source = new Position2D;
		_receiver = new Position2D;
		_platformEnd = new Position2D;
		_O1 = new Position2D;
		_O2 = new Position2D;
		_sourceImage = new Position2D;

		// Fill and check data
		if(freq <= 0)
		{
			printf (".ERROR: frequency = 0 \n") ;
			throw new int( ERRFrequency );
		}

		if (!DataInPerpendicularPlane(terrainItems))
		{
			return;
		}

		if (!CheckData(terrainItems))
		{
			return;
		}

		// source image S''
		double aSlope = 0;
		double bSlope = 0;
		bool verticalSlope = LineCoeff(_O1, _O2, aSlope, bSlope);
		SourceImage(aSlope, bSlope, verticalSlope);
		if(_receiver->d == _sourceImage->d && _receiver->z == _sourceImage->z)
		{
			return;
		}

		// Intersection between (S''R) and (O1O2)
		double aSimageR = 0;
		double bSimageR = 0;
		bool verticalSimageR = LineCoeff(_sourceImage, _receiver, aSimageR, bSimageR);
		Position2D intersect;
		if(!LinesIntersection(aSlope, bSlope, verticalSlope, aSimageR, bSimageR, verticalSimageR, &intersect))
		{
			return;
		}
		
		// Attenuation calculation
		EmbankmentCalculation(&intersect, freq);
	}

	/*!
	 * \brief Calculates embankment attenuation (in _embankmentAttenuation)
	 * 
	 * \param intersect The intersection point (I)
	 * 
	 * \param freq The frequency
	 */
	void Embankment::EmbankmentCalculation(Position2D const* intersect, double freq)
	{
		// Half-length e 
		//  9.3.5 formula (28) p.43
		double angleTheta = ThetaAngle();
		double cosAngleTheta = cos(angleTheta);
		if(cosAngleTheta == 0)
		{
			return;
		}
		double SimageI = distance2D(_sourceImage, intersect);
		double IR = distance2D(intersect, _receiver); 
		if(SimageI == 0 && IR == 0)
		{
			return;
		}

		double halfLength = HalfLength(cosAngleTheta, freq, SimageI, IR);

		// I x-coordinate on (O1O2) :
		double xI = distance2D(_O1, intersect);
		double prodScal = (_O2->d - _O1->d) * (intersect->d - _O1->d) + (_O2->z - _O1->z) * (intersect->z - _O1->z);
		if (prodScal < 0)
		{
			xI = -1 * xI;
		}

		// Epsilon (formula (29) p. 44)
		double lT = distance2D(_O1, _O2);
		double epsilon = 0;
		if(xI > (-1 * halfLength) && xI < (lT + halfLength))
		{
			double val1 = lT / (2 * halfLength);
			double val2 = (xI - halfLength) / (2 * halfLength);
			double val3 = (xI + halfLength) / (2 * halfLength);
			if (val1 <= val3) //(val1 <= val2) //// 
			{
				epsilon = val1;
			}
			else
			{
				epsilon = val3; //// val2; 
			}

			if(val2 > 0) //// 
			{
				epsilon -= val2;
			}
		}

		// Attenuation (formula (30) p.44)
		_embankmentAttenuation = -1.5 * epsilon * (2 - _Gembankment);
	}

	/*!
	 * \brief Calculates the 2D position in the perpendicular plane to road
	 * 
	 * \param posToFill The position the fill
	 * 
	 * \param posToRead The position to read
	 * 
	 * \param cosTheta cos(Theta) 
	 */
	void Embankment::FillPosInPerpendicularPlane(Position2D* posToFill, Position2D const* posToRead, double cosTheta)
	{
		if(posToFill == NULL || posToRead == NULL || cosTheta > 1 || cosTheta < -1)
		{
			return;
		}

		posToFill->d = posToRead->d * cosTheta;
		posToFill->z = posToRead->z;
	}

	/*!
	 * \brief Calculates the 2D position of needed items in the perpendicular plane to road
	 *	and check data
	 * 
	 * \param terrainItems The terrain items
	 * 
	 * \returns True if the data are OK
	 * 
	 * \throws ERREmbankment
	 */
	bool Embankment::DataInPerpendicularPlane(vector<ProfilePointNMPB*> terrainItems)
	{
		int nbItems = (int) terrainItems.size();
		if (nbItems < 3)
		{
			return false;
		}

		ProfilePointNMPB* source = terrainItems[0];
		if(source->ext == NULL || source->ext->type != ETRoadSource_NMPB)
		{
			return false;
		}

		double cosTheta = source->ext->cosTheta;

		if(_source) delete _source;
		_source = new Position2D;
		Position2D posSource = source->position2D_withHeight();
		FillPosInPerpendicularPlane(_source, &posSource, cosTheta);
		bool embankmentFound = false;
		bool platformFound = false;
		for (int i = 1 ; i < nbItems ; i ++)
		{
			ProfilePointNMPB* terrainItem = terrainItems[i];		

			if(terrainItem->ext != NULL) 
			{
				if(terrainItem->ext->type == ETPlatform_NMPB)
				{
					// platform only before embankment
					if(!embankmentFound)
					{
						if(_platformEnd) delete _platformEnd;
						_platformEnd = new Position2D;
						FillPosInPerpendicularPlane(_platformEnd, &terrainItem->position2D, cosTheta);
					}
				}
				else if(terrainItem->ext->type == ETEmbankment_NMPB)
				{
					if(embankmentFound == true)
					{
						// only one embankment is accepted
						printf (".ERROR: 2 embankments detected \n") ;
						throw new int( ERREmbankment );
					}
					if(_O1) delete _O1;
					_O1 = new Position2D;
					FillPosInPerpendicularPlane(_O1, &terrainItems[i-1]->position2D, cosTheta);
					platformFound = true;

					if(_O2) delete _O2;
					_O2 = new Position2D;
					FillPosInPerpendicularPlane(_O2, &terrainItem->position2D, cosTheta);
					_Gembankment = terrainItem->impedance;
					embankmentFound = true;
				}
			}
		}

		if(!embankmentFound || !platformFound)
		{
			return false;
		}

		if(_receiver) delete _receiver;
		_receiver = new Position2D;
		Position2D lastPos = terrainItems[nbItems - 1]->position2D_withHeight();
		FillPosInPerpendicularPlane(_receiver, &lastPos, cosTheta);

		// Calculation of _angleAlpha
		if(_O1->d == _O2->d && _O1->z == _O2->z)
		{
			return false;
		}
		_angleAlpha = atan2(_O2->z - _O1->z, _O2->d - _O1->d);

		return true;
	}

	/*!
	* \brief Checks the data
	*
	*	Checking according  9.3.5 p. 43
	*
	* \param terrainItems The terrain items
	* 
	* \returns true if embankment can be calculated
	*/
	bool Embankment::CheckData(vector<ProfilePointNMPB*> terrainItems)
	{
		// check slope :
		if (_angleAlpha < PI / 12 || _angleAlpha > PI / 4)
		{
			return false;
		}

		// the foot of the bank and the receiver are in direct view :
		double aCoef = 0;
		double bCoef = 0;
		bool vert = LineCoeff(_O1, _receiver, aCoef, bCoef);
		if(!vert)
		{
			for (int i = 0 ; i < (int) terrainItems.size() ; i++)
			{ 
				ProfilePointNMPB* elem = terrainItems[i];
				Position2D posElem = elem->position2D_withHeight();
				if(posElem.d > _O1->d && posElem.d < _receiver->d)
				{
					if(posElem.z >= aCoef * posElem.d + bCoef)
					{
						return false;
					}
				}
			}
		}

		// the foot of the bank is at a maximum distance of 10 metres from the edge of the closest platform
		double distToPlatform = distance2D(_platformEnd, _O1);
		if(distToPlatform > 10)
		{
			return false;
		}

		// the foot of the bank is lower than or at the same height as the edge of the closest platform
		if(_O1->z > _O2->z)
		{
			return false;
		}

		return true;
	}

	/*!
	 * \brief Calculates line coefficients : z = a*d + b (or d = b if line vertical)
	 *	Warning The 2 positions must be different
	 *
	 * \param pos1 The first line point 
	 * 
	 * \param pos2 The second line pont
	 * 
	 * \param aCoef The a coefficient to fill
	 * 
	 * \param bCoef The b coefficient to fill
	 * 
	 * \returns true if the line is vertical
	 */
	bool Embankment::LineCoeff(Position2D const* pos1, Position2D const* pos2, double &aCoef, double &bCoef)
	{
		double dividende = pos2->d - pos1->d ;
		if(dividende != 0)
		{
			// z = a*d + b
			aCoef = (pos2->z - pos1->z) / dividende;
			bCoef = pos1->z - aCoef * pos1->d;
			return false;
		}
		else
		{
			// d = b
			aCoef = 0;
			bCoef = pos1->d;
			return true;
		}
	}

	/*!
	 * \brief Calculates the source image by the line z = a * d + b (or d = b if vertical)
	 * 
	 * \param aCoef The a coefficient
	 * 
	 * \param bCoef The b coefficient
	 * 
	 * \param verticalSlope true if vertical line
	 */
	void Embankment::SourceImage(double aCoef, double bCoef, bool verticalSlope)
	{
		if(_sourceImage) delete _sourceImage;
		_sourceImage = new Position2D;
		if(verticalSlope)
		{
			// d = bCoef
			_sourceImage->d = 2 * bCoef - _source->d;
			_sourceImage->z = _source->z;

		}	
		else
		{
			// z = aCoef * d + bCoef
			double dividende = aCoef * aCoef + 1;
			_sourceImage->d = (-1 * aCoef * aCoef * _source->d + 2 * aCoef * _source->z + _source->d - 2 * aCoef * bCoef) / dividende;
			_sourceImage->z = (aCoef * aCoef * _source->z + 2 * aCoef * _source->d - _source->z + 2 * bCoef) / dividende;
		}
	}

	/*!
	 * \brief Calculates the lines intersection
	 * 
	 * \param a1 Parameter a1 of the first line
	 * 
	 * \param b1 Parameter b1 of the first line
	 * 
	 * \param vert1 True if the first line is vertical
	 * 
	 * \param a2 Parameter a2 of the second line
	 * 
	 * \param b2 Parameter b2 of the second line
	 * 
	 * \param vert2 True if the second line is vertical
	 * 
	 * \param Ipoint The intersection point to fill
	 * 
	 * \returns true if there is an only intersection
	 */
	bool Embankment::LinesIntersection(double a1, double b1, bool vert1, double a2, double b2, bool vert2, Position2D* Ipoint)
	{
		if(Ipoint == NULL)
		{
			return false;
		}
		if (vert1)
		{
			if(vert2)
			{
				// parallel lines : 0 or infinity intersections
				return false;
			}
			else
			{
				Ipoint->d = b1;
				Ipoint->z = a2 * b1 + b2;
			}
		}
		else
		{
			if(vert2)
			{
				Ipoint->d = b2;
				Ipoint->z = a1 * b2 + b1;
			}
			else
			{
				double dividende = a2 - a1;
				if(dividende == 0)
				{
					// parallel lines : 0 or infinity intersections
					return false;
				}

				Ipoint->d = (b1 - b2) / dividende;
				Ipoint->z = a1 * Ipoint->d + b1;
			}
		}

		return true;
	}

	/*!
	 * \brief Calculates the Theta angle in the formula (28) p.43
	 * 
	 * \returns the angle
	 */
	double Embankment::ThetaAngle()
	{
		// NB : _angleAlpha already calculated when check data
		
		// the (O1x,S''R) angle
		double angleBeta = atan2(_receiver->z - _sourceImage->z, _receiver->d - _sourceImage->d);
		
		return _angleAlpha + angleBeta;
	}

	/*!
	 * \brief Calculates half-length e
	 * 
	 * p. 43 -  9.3.5 formula (28)
	 *
	 * \param cosAngleTheta The theta cosines (must be != 0)
	 * 
	 * \param freq The frequency (must be > 0)
	 * 
	 * \param SimageI The distance between Simage and I
	 * 
	 * \param IR The distance between Receiver and I
	 * 
	 * \returns the half-length e
	 */
	double Embankment::HalfLength(double cosAngleTheta, double freq, double SimageI, double IR)
	{
		// Already checked : freq > 0, cosAngleTheta != 0, SimageI + IR != 0
		double wavelength =  cSound / freq;
		double val1 = (wavelength / 2) * (SimageI * IR)/(SimageI + IR) + (wavelength * wavelength / 16);
		double halfLength = sqrt(val1) / cosAngleTheta;

		return halfLength;
	}

}

