/*!
* \file CalculRoadEmission.cpp
* \brief Calculations of road emissions
* \author CSTB
* \version 1.0
*/
#include "CalculRoadEmission.h"
#include <stdio.h>

namespace CalculRoadEmissionNMPB
{

	// ------------------------------ Global variables and constants ---------------------------------------

	/*!
	* \brief Frequencies in third octave band
	*/
	const int ThirdOctaveFrequencies[] = {100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000};
	/*!
	* \brief Frequencies in octave band
	*/
	const int OctaveFrequencies[] = {125, 250, 500, 1000, 2000, 4000};
	/*!
	* \brief Just the global value in dB(A)
	*/
	const int AllFrequencies[] = {0};

	/*!
	* \brief frequencies maximal number
	*/
	const int frequencyNumber = 18;
	/*!
	* \brief Nominal median frequencies array
	*/
	//const int NominalMedianFrequency[] = {100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000};
	//const int* NominalMedianFrequency = ThirdOctaveFrequencies ;
#define NominalMedianFrequency ThirdOctaveFrequencies
	/*!
	* \brief spectral distribution for draining ground ( 2.8, array (2.23), p. 29)
	*/
	const double spectralDistribution_draining[] =	{-22, -22, -20, -17, -15, -12, -10, -8, -9, -9, -10, -11, -12, -13, -16, -18, -20, -23};
	/*!
	* \brief spectral distribution for non draining ground ( 2.8, array (2.23), p. 29)
	*/
	const double spectralDistribution_nonDraining[] = {-27, -26, -24, -21, -19, -16, -14, -11, -11, -8, -7, -8, -10, -13, -16, -18, -21, -23};
	/*!
	* \brief Filter A corrections
	*/
	const double spectralFilterA[] = {-19.1, -16.1, -13.4, -10.9, -8.6, -6.6, -4.8, -3.2, -1.9, -0.8, 0.0, 0.6, 1.0, 1.2, 1.3, 1.2, 1.0, 0.5};


	/*!
	* \brief Road surface descriptions by id (RoadSurfaceType) : global variable
	*/
	map<int,RoadSurfaceDescription> RoadSurfaceDescriptionMap;


	// ---------------------------- Road surface definition  ---------------------------

	/*!
	* \brief Definition of a new road surface type with data needed for rolling component and for spectrum
	*
	* \param idRoadSurface The road surface id
	*
	* \param AVL The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	*
	* \param BVL The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	*
	* \param Vref_VL The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	*
	* \param APL The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	*
	* \param BPL The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	*
	* \param Vref_PL The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	*
	* \param spectrum The spectral distribution for this road surface (if null, non draining default values will be used)
	*
	* \returns true if all OK
	*/
	bool DefineRoadSurfaceType(int idRoadSurface,
		double AVL, double BVL, double Vref_VL,
		double APL, double BPL, double Vref_PL,
		double const* spectrum)
	{
		if(idRoadSurface < RoadSurface_UserDefined)
		{
			printf (".ERROR: definition of road surface type with idRoadSurface < RoadSurface_UserDefined \n") ;
			return false;
		}

		double const* spectrumToUse;
		bool doNormalization = true;
		if(spectrum == NULL)
		{
			spectrumToUse = spectralDistribution_nonDraining;
			doNormalization = false;
		}
		else
		{
			spectrumToUse = spectrum;
		}

		RoadSurfaceDescription roadSurfDesc(AVL, BVL, Vref_VL, APL, BPL, Vref_PL, spectrumToUse, doNormalization);
		RoadSurfaceDescriptionMap[idRoadSurface] = roadSurfDesc;
		return true;
	}


	// ------------------------------ CalculRoadEmission Class ---------------------------------------

	/*!
	 * \brief If RoadSurfaceType < RoadSurface_UserDefined, add it to the map (if not already in)
	 *
	 * \param surfType The surface type id
	 */
	void CalculRoadEmission::FillRoadSurfaceDescriptionMap(RoadSurfaceType surfType)
	{
		if(surfType < RoadSurface_UserDefined && RoadSurfaceDescriptionMap.find(static_cast<int>(surfType)) == RoadSurfaceDescriptionMap.end())
		{
			double AVL = 0;
			double BVL = 0;
			double Vref_VL = 90;
			double APL = 0;
			double BPL = 0;
			double Vref_PL = 80;
			const double* spectrum = spectralDistribution_nonDraining;

			// data from array 2.6 p.21 -  2.7.2.2
			switch(surfType)
			{
			case RoadSurface_R1:
			case RoadSurface_DR1:
				AVL = 21.0;
				BVL = 53.4;
				APL = 20.0;
				BPL = 61.5;
				break;
				/*
				AVL = 20.1;
				BVL = 55.4;
				APL = 20.0;
				BPL = 63.4;
				break;
				*/
			case RoadSurface_R3:
			case RoadSurface_DR3:
				AVL = 21.4;
				BVL = 57.5;
				APL = 20.0;
				BPL = 64.1; // modifi le 19/09/2012, -0.1 dB par rapport aux versions prcdentes
				break;
			case RoadSurface_R2:
			case RoadSurface_DR2:
			default:
				// like R2
				AVL = 20.1;
				BVL = 55.4;
				APL = 20.0;
				BPL = 63.3; // modifi le 19/09/2012, -0.1 dB par rapport aux versions prcdentes
				break;
			}

			if(surfType == RoadSurface_DR1 || surfType == RoadSurface_DR2 || surfType == RoadSurface_DR3)
			{
				spectrum = spectralDistribution_draining;
			}

			RoadSurfaceDescription roadSurfDesc(AVL, BVL, Vref_VL, APL, BPL, Vref_PL, spectrum, false);
			RoadSurfaceDescriptionMap[surfType] = roadSurfDesc;
		}
	}

	/*!
	* \brief : Fill the linear corrections map
	*
	*/

	void CalculRoadEmission::FillLinearCorrectionMap()
	{
		_linearCorrection.clear();
		for(int i = 0; i < frequencyNumber ; i++)
		{
		 // Modification
		 // DVM 18/11/2011
		 // Correction for removing dB(A) filtering is equal to minus the value of the filter
		 // _linearCorrection[NominalMedianFrequency[i]] = spectralFilterA[i];
			_linearCorrection[NominalMedianFrequency[i]] = -spectralFilterA[i];
		}
	}

	/*!
	* \brief Give spectral distribution of the power sound value
	*
	* \param powerValue The sound power value
	*
	* \param surfaceType The road surface type
	*
	* \returns an array containing the sound value for each frequency band
	*/
	double* CalculRoadEmission::SpectralDistribution(double powerValue, RoadSurfaceType surfaceType)
	{
		double* specResults = _Lw_band ;

		for(int i = 0 ; i < _bandsNumber ; i++)
		{
			double valToAdd = 0;

			// spectral distribution :
			if(RoadSurfaceDescriptionMap.find(surfaceType) != RoadSurfaceDescriptionMap.end())
			{
				RoadSurfaceDescription roadSurfDesc = RoadSurfaceDescriptionMap[surfaceType];
				map<int,double> specDistrMap = roadSurfDesc.get_spectralDistributionMap();
				if(specDistrMap.find(_usedFrequencies[i])!= specDistrMap.end())
				{
					valToAdd = specDistrMap[_usedFrequencies[i]];
				}
			}

			// linear correction :
			if(!_specA && _linearCorrection.find(_usedFrequencies[i]) != _linearCorrection.end())
			{
				valToAdd += _linearCorrection[_usedFrequencies[i]];
			}

			specResults[i] = powerValue + valToAdd;
		}

		return specResults;
	}


	/*!
	 * \brief Initialization of CalculRoadEmission class
	 */
	CalculRoadEmission::CalculRoadEmission(void)
	{
		_Lw_VL = 0;
		_Lw_PL = 0;
		_QVL = 0;
		_QPL = 0;
		_Lw_rolling_VL = 0;
		_Lw_rolling_PL = 0;
		_Lw_traction_VL = 0;
		_Lw_traction_PL = 0;

		// if SpectrumType non specified, third octave band by default
		_bandsNumber = 18;
		_Lw_band = 0 ;
		_usedFrequencies = ThirdOctaveFrequencies;
		_specA = true;
		FillLinearCorrectionMap();
	}

	/*!
	* \brief Initialization of CalculRoadEmission class with given SpectrumType and the given surface type
	*
	* \param specType The spectrum type
	*
	* \param surfType The road surface type
	*/
	CalculRoadEmission::CalculRoadEmission(RoadSpectrumType specType, RoadSurfaceType surfType)
	{
		_Lw_VL = 0;
		_Lw_PL = 0;
		_QVL = 0;
		_QPL = 0;
		_Lw_rolling_VL = 0;
		_Lw_rolling_PL = 0;
		_Lw_traction_VL = 0;
		_Lw_traction_PL = 0;

		_bandsNumber = 0;
		switch(specType)
		{
		case Spectrum_dBA:
			_bandsNumber = 1;
			_specA = true;
			_usedFrequencies = AllFrequencies;
			break;
		case Spectrum_oct_A:
			_bandsNumber = 6;
			_specA = true;
			_usedFrequencies = OctaveFrequencies;
			break;
		case Spectrum_3oct_A:
			_bandsNumber = 18;
			_specA = true;
			_usedFrequencies = ThirdOctaveFrequencies;
			break;
		case Spectrum_oct_lin:
			_bandsNumber = 6;
			_specA = false;
			_usedFrequencies = OctaveFrequencies;
			break;
		case Spectrum_3oct_lin:
			_bandsNumber = 18;
			_specA = false;
			_usedFrequencies = ThirdOctaveFrequencies;
			break;
		default:
			_bandsNumber = 18;
			_specA = true;
			_usedFrequencies = ThirdOctaveFrequencies;
			break;
		}

		_Lw_band = 0 ;

		FillRoadSurfaceDescriptionMap(surfType);
		FillLinearCorrectionMap();
	}

	/*!
	 * \brief Calculation of the Sound power level per source line meter
	 *
	 * \param roadTraffic The road traffic data
	 *
	 * \returns the Sound power level per source line meter, for each third-octave band
	 */
	double* CalculRoadEmission::SoundPowerLevelPerMeter (RoadTraffic* roadTraffic)
	{
		if(roadTraffic == NULL)
		{
			printf("Erreur : le composant RoadTraffic est nul \n");
			return 0;
		}

		double Lw_total = 0 ;
		for(int i = 0 ; i < roadTraffic->nbComponents ; i++)
		{
			RoadTrafficComponent rtComponent = roadTraffic->traffic[i];
			if (rtComponent.vehicleType == VehicleType_VL)
			{
				switch (rtComponent.flowType)
				{
				case FlowType_CONST :
				case FlowType_ACC :
				case FlowType_DEC :
					{
						_Lw_rolling_VL = Lw_rolling_VL (rtComponent.trafficSpeed,
												        roadTraffic->surfaceType,
														roadTraffic->surfaceAge);
						_Lw_traction_VL = Lw_traction_VL (rtComponent.trafficSpeed,
														  rtComponent.flowType);
						//  2.7.1.1, formula (19), p. 19 :
						_Lw_VL = EnergeticAddition(_Lw_rolling_VL, _Lw_traction_VL);
						break ;
					}
				case FlowType_START :
				case FlowType_STOP :
					{
						_Lw_VL = Lw_startStop (rtComponent.vehicleType,
											   roadTraffic->ramp,
											   rtComponent.flowType,
											   roadTraffic->surfaceAge);
						FillRollingAndTractionFromStartStop(_Lw_VL, rtComponent.vehicleType);
						break ;
					}
				default:
					printf ("ERREUR : FlowType inconnu \n") ;
					return 0 ;
				}
				_QVL = rtComponent.trafficFlow;
				Lw_total += _QVL * pow (10., _Lw_VL / 10.) ;
			}
			else if (rtComponent.vehicleType == VehicleType_PL)
			{
				switch (rtComponent.flowType)
				{
				case FlowType_CONST :
				case FlowType_ACC :
				case FlowType_DEC :
					{
						_Lw_rolling_PL = Lw_rolling_PL (rtComponent.trafficSpeed,
													    roadTraffic->surfaceType,
														roadTraffic->surfaceAge);
						_Lw_traction_PL = Lw_traction_PL (rtComponent.trafficSpeed,
														  rtComponent.flowType,
														  roadTraffic->ramp);
						//  2.7.1.1, formula (19), p. 19 :
						_Lw_PL = EnergeticAddition(_Lw_rolling_PL, _Lw_traction_PL);
						break ;
					}
				case FlowType_START :
				case FlowType_STOP :
					{
						_Lw_PL = Lw_startStop (rtComponent.vehicleType,
										       roadTraffic->ramp,
											   rtComponent.flowType,
											   roadTraffic->surfaceAge);
						FillRollingAndTractionFromStartStop(_Lw_PL, rtComponent.vehicleType);
						break ;
					}
				default:
					printf ("ERREUR : FlowType inconnu \n") ;
					return 0 ;
				}
				_QPL = rtComponent.trafficFlow;
				Lw_total += _QPL * pow (10., _Lw_PL / 10.) ;
			}
		}

		// sum on vehicles :
		// double Lw = LwTotal(_Lw_VL, _QVL, _Lw_PL, _QPL);
		double Lw = (Lw_total > 0) ? 10. * log10(Lw_total) : 0.0 ;

		// by frequency bands :
		_Lw_band = roadTraffic->_Lw_values ;
		return SpectralDistribution(Lw, roadTraffic->surfaceType);
	}

	/*!
	 * \brief Calculation of the Sound power level for Start or Stop section
	 *
	 *  2.7.4, array (2.22-A) and (2.22-B), p. 29
	 *
	 * \param vtype The vehicle type
	 *
	 * \param ramp The ramp
	 *
	 * \param flowType The flow type
	 *
	 * \param surfaceAge The surface age
	 *
	 * \returns the calculated sound power level for start or stop section
	 */
	double CalculRoadEmission::Lw_startStop (RoadVehicleType vtype,
										     double ramp,
											 RoadFlowType flowType,
											 double surfaceAge)
	{
		double lwStartStop_new = 0 ;
		double lwStartStop_old = 0 ;

		// linear interpolation as a function of surface age
		// weight = 1 if age < 2
		// weight = 0 if age > 10

		double weightAge = 1 ;
		if (surfaceAge > 10)
			weightAge = 0 ;
		else if (surfaceAge > 2)
			weightAge = (10 - surfaceAge) / 8 ;

		if (vtype == VehicleType_VL)
		{
			if (flowType == FlowType_START)
			{
				lwStartStop_new = 51.1 ;
				lwStartStop_old = 51.3 ;
			}
			else
			{
				lwStartStop_new = 44.5 ;
				lwStartStop_old = 45.1;
			}
		}
		else
		{
			// (vtype == VehicleType_PL)

			if (flowType == FlowType_START)
			{
				lwStartStop_new = 62.4 ;
				lwStartStop_old = 62.5;

				if (ramp > 4.5 && ramp <= 6)
				{
					lwStartStop_new += 2 * (ramp - 4.5);
					lwStartStop_old += 2 * (ramp - 4.5);
				}
			}
			else
			{
				// (flowType == FlowType_STOP)

				lwStartStop_new  = 58.0 ;
				lwStartStop_old  = 58.3 ;

				if(ramp >= -6 && ramp <= -2)
				{
					lwStartStop_new += (ramp + 2) ; // ramp < 0 !
					lwStartStop_old += (ramp + 2) ; // ramp < 0 !
				}
			}
		}

		double lwStartStop = weightAge * lwStartStop_new + (1 - weightAge) * lwStartStop_old ;

		return lwStartStop;
	}

	/*!
	 * \brief Fill the 2 components _Lw_rolling_veh and _Lw_traction_veh from the calculated Sound power level for Start or Stop section
	 *
	 * \param lwStartStop The calculated Sound power level for Start or Stop section
	 *
	 * \param vtype The vehicle type
	 */
	void CalculRoadEmission::FillRollingAndTractionFromStartStop (double lwStartStop,
																  RoadVehicleType vtype)
	{
		// assume Lw is divided in 2 equal parts (rolling and traction), so :
		// _Lw_rolling = _Lw_traction = lwStartStop - 10 * log(2)
		double meanLw = lwStartStop - 10 * log10(2.);
		if(vtype == VehicleType_VL)
		{
			_Lw_rolling_VL = meanLw;
			_Lw_traction_VL = meanLw;
		}
		else
		{
			_Lw_rolling_PL = meanLw;
			_Lw_traction_PL = meanLw;
		}
	}

	/*!
	 * \brief Calculation of the Sound power level per source line meter with VL and PL data
	 *
	 *  2.3.5, formula (8), p. 13 (without spectral distribution)
	 *
	 * \param LwVL The Sound power level for VL
	 *
	 * \param QVL : average hourly traffic flow for VL
	 *
	 * \param LwPL The Sound power level for PL
	 *
	 * \param QPL : average hourly traffic flow for PL
	 *
	 * \returns the total sound power level
	 */
	double CalculRoadEmission::LwTotal(double LwVL, double QVL, double LwPL, double QPL)
	{
		double WVL = QVL * pow (10., LwVL/10) ;
		double WPL = QPL * pow (10., LwPL/10) ;
		double WTot = WVL + WPL ;
		double Lw = (WTot > 0) ? 10 * log10(WTot) : 0.0 ;
		return Lw;
	}

	/*!
	* \brief Energetic addition
	*
	*  2.7.1.1, p. 19
	*
	* \param val1 The first value to add
	*
	* \param val2 The second value to add
	*
	* \returns The energetic addition of val1 and val2
	*/
	double CalculRoadEmission::EnergeticAddition(double val1, double val2)
	{
		return 10.0 * log10(pow(10.,val1 / 10.) + pow(10.,val2 / 10.));
	}


	// ---------------------------- Rolling component  ---------------------------

	/*!
	* \brief Calculation of the rolling component of the Sound power level per source line meter
	*
	* \param roadTraffic The road traffic data
	*
	* \returns The rolling component of the Sound power level per source line meter, for each third-octave band
	*/
	double* CalculRoadEmission::Lwm_rolling (RoadTraffic* roadTraffic)
	{
		if(roadTraffic == NULL)
		{
			printf("Erreur : le composant RoadTraffic est nul \n");
			return 0;
		}

		for(int i = 0 ; i < roadTraffic->nbComponents ; i++)
		{
			RoadTrafficComponent rtComponent = roadTraffic->traffic[i];
			if (rtComponent.vehicleType == VehicleType_VL)
			{
				if(rtComponent.flowType != FlowType_START && rtComponent.flowType != FlowType_STOP)
				{
					_Lw_rolling_VL = Lw_rolling_VL (rtComponent.trafficSpeed, roadTraffic->surfaceType, roadTraffic->surfaceAge);
				}
				else
				{
					double lw = Lw_startStop(rtComponent.vehicleType, roadTraffic->ramp, rtComponent.flowType, roadTraffic->surfaceAge);
					FillRollingAndTractionFromStartStop(lw, rtComponent.vehicleType);
				}

				_QVL = rtComponent.trafficFlow;
			}
			else if (rtComponent.vehicleType == VehicleType_PL)
			{
				if(rtComponent.flowType != FlowType_START && rtComponent.flowType != FlowType_STOP)
				{
					_Lw_rolling_PL = Lw_rolling_PL (rtComponent.trafficSpeed, roadTraffic->surfaceType, roadTraffic->surfaceAge);
				}
				else
				{
					double lw = Lw_startStop(rtComponent.vehicleType, roadTraffic->ramp, rtComponent.flowType, roadTraffic->surfaceAge);
					FillRollingAndTractionFromStartStop(lw, rtComponent.vehicleType);
				}

				_QPL = rtComponent.trafficFlow;
			}
		}

		// sum on vehicles :
		double Lw = LwTotal(_Lw_rolling_VL, _QVL, _Lw_rolling_PL, _QPL);

		// by frequency bands :
		_Lw_band = roadTraffic->_Lw_values ;
		return SpectralDistribution(Lw, roadTraffic->surfaceType);
	}

	/*!
	* \brief Calculation of the rolling component of the Sound power level per source line meter for a vehicle type
	*
	* \param vtype The vehicle type
	*
	* \param vehicleSpeed The vehicle speed
	*
	* \param surfaceType The surface type
	*
	* \param surfaceAge The surface age
	*
	* \returns the sound power associated with the rolling component of a single vehicle
	*/
	double* CalculRoadEmission::Lw_rolling (RoadVehicleType vtype,
											double vehicleSpeed,
											RoadSurfaceType surfaceType,
											double surfaceAge)
	{
		double Lr_w_veh = 0;
		switch(vtype)
		{
		case VehicleType_VL :
			Lr_w_veh = Lw_rolling_VL(vehicleSpeed, surfaceType, surfaceAge);
			_Lw_rolling_VL = Lr_w_veh;
			break;
		case VehicleType_PL:
			Lr_w_veh = Lw_rolling_PL(vehicleSpeed, surfaceType, surfaceAge);
			_Lw_rolling_PL = Lr_w_veh;
			break;
		default:
			Lr_w_veh = 0;
			printf("Erreur lors du calcul de la composante roulement : type de vehicule non reconnu \n");
			break;
		}

		// Annexe 4, p.113
		// Correction DVM 09/10/2013: inversion des signes, les fonctions lmentaires du document SETRA
		// permettent de calculer un niveau de pussance par mtre (Lw/m) pour 1 vh/h... cette fonction
		// retourne la puissance d'un vhicle isol.
		Lr_w_veh = Lr_w_veh + 10 * log10(vehicleSpeed) + 30;

		_Lw_band = _output_values ;
		return SpectralDistribution(Lr_w_veh, surfaceType);
	}

	/*!
	 * \brief Rolling component of the Sound power level for VL
	 *
	 *  2.7.2, arrays (2.6) and (2.8), p. 19-22
	 *
	 * \param vehicleSpeed The vehicle speed
	 *
	 * \param surfaceType The surface type
	 *
	 * \param surfaceAge The surface age
	 *
	 * \returns the rolling component of the Sound power level per source line meter, for VL
	 */
	double CalculRoadEmission::Lw_rolling_VL (double vehicleSpeed, RoadSurfaceType surfaceType, double surfaceAge)
	{
		if(vehicleSpeed < 20 || vehicleSpeed > 130)
		{
			printf("Erreur lors du calcul de la composante roulement : la vitesse %.2f, utilisee pour un vehicule leger, n'est pas comprise entre 20 et 130 km/h \n", vehicleSpeed);
			return 0;
		}

		double Lr_w_veh = 0;

		//  2.7.2.2, array (2.6) p. 21
		RoadSurfaceDescription roadSurfdescr = RoadSurfaceDescriptionMap[surfaceType];
		double Vref = 90;
		if(roadSurfdescr.get_Vref_VL() != 0) Vref = roadSurfdescr.get_Vref_VL();
		Lr_w_veh = roadSurfdescr.get_BVL() + roadSurfdescr.get_AVL() * log10(vehicleSpeed / Vref);

		// Age for R1, R2 or R3 :  2.7.2.3, array (2.8) p. 22
		switch(surfaceType)
		{
		case RoadSurface_R1:
		case RoadSurface_DR1:
			if (surfaceAge <= 2)
			{
				Lr_w_veh += -4;
			}
			else if(surfaceAge <= 10)
			{
				Lr_w_veh += 0.5 * (surfaceAge - 10);
			}
			break;

		case RoadSurface_R2:
		case RoadSurface_DR2:
			if (surfaceAge <= 2)
			{
				Lr_w_veh += -2;
			}
			else if(surfaceAge <= 10)
			{
				Lr_w_veh += 0.25 * (surfaceAge - 10);
			}
			break;

		case RoadSurface_R3:
		case RoadSurface_DR3:
			if (surfaceAge <= 2)
			{
				Lr_w_veh += -1.6;
			}
			else if(surfaceAge <= 10)
			{
				Lr_w_veh += 0.2 * (surfaceAge - 10);
			}
			break;
		default:
			break;
		}

		return Lr_w_veh;
	}

	/*!
	* \brief Rolling component of the Sound power level for PL
	*
	*  2.7.2, arrays (2.6) and (2.8), p. 19-22
	*
	* \param vehicleSpeed The vehicle speed
	*
	* \param surfaceType The surface type
	*
	* \param surfaceAge The surface age
	*
	* \returns the rolling component of the Sound power level per source line meter, for PL
	*/
	double CalculRoadEmission::Lw_rolling_PL (double vehicleSpeed, RoadSurfaceType surfaceType, double surfaceAge)
	{
		if(vehicleSpeed < 20 || vehicleSpeed > 100)
		{
			printf("Erreur lors du calcul de la composante roulement : la vitesse %.2f, utilisee pour un poids lourd, n'est pas comprise entre 20 et 100 km/h \n", vehicleSpeed);
			return 0;
		}

		double Lr_w_veh = 0;

		//  2.7.2.2, array (2.6) p. 21
		RoadSurfaceDescription roadSurfdescr = RoadSurfaceDescriptionMap[surfaceType];
		double Vref = 80;
		if(roadSurfdescr.get_Vref_PL() != 0) Vref = roadSurfdescr.get_Vref_PL();
		Lr_w_veh = roadSurfdescr.get_BPL() + roadSurfdescr.get_APL() * log10(vehicleSpeed / Vref);

		//  Age for R1, R2 or R3 :  2.7.2.3, array (2.8) p. 22
		switch(surfaceType)
		{
		case RoadSurface_R1:
		case RoadSurface_DR1:
			if (surfaceAge <= 2)
			{
				Lr_w_veh += -2.4;
			}
			else if(surfaceAge <= 10)
			{
				Lr_w_veh += 0.3 * (surfaceAge - 10);
			}
			break;

		case RoadSurface_R2:
		case RoadSurface_DR2:
			if (surfaceAge <= 2)
			{
				Lr_w_veh += -1.2;
			}
			else if(surfaceAge <= 10)
			{
				Lr_w_veh += 0.15 * (surfaceAge - 10);
			}
			break;

		case RoadSurface_R3:
		case RoadSurface_DR3:
			if (surfaceAge <= 2)
			{
				Lr_w_veh += -1.;
			}
			else if(surfaceAge <= 10)
			{
				Lr_w_veh += 0.12 * (surfaceAge - 10);
			}
			break;
		default:
			break;
		}

		return Lr_w_veh;
	}


	// ---------------------------- Traction component  ---------------------------

	/*!
	* \brief Calculation of the traction component of the Sound power level per source line meter
	*
	* \param roadTraffic The road traffic data
	*
	* \returns the traction component of the Sound power level per source line meter, for each third-octave band
	*/
	double* CalculRoadEmission::Lwm_traction (RoadTraffic* roadTraffic)
	{
		if(roadTraffic == NULL)
		{
			printf("Erreur : le composant RoadTraffic est nul \n");
			return 0;
		}

		for(int i = 0 ; i < roadTraffic->nbComponents ; i++)
		{
			RoadTrafficComponent rtComponent = roadTraffic->traffic[i];
			if (rtComponent.vehicleType == VehicleType_VL)
			{
				if(rtComponent.flowType != FlowType_START && rtComponent.flowType != FlowType_STOP)
				{
					_Lw_traction_VL = Lw_traction_VL(rtComponent.trafficSpeed, rtComponent.flowType);
				}
				else
				{
					double lw = Lw_startStop(rtComponent.vehicleType, roadTraffic->ramp, rtComponent.flowType, roadTraffic->surfaceAge);
					FillRollingAndTractionFromStartStop(lw, rtComponent.vehicleType);
				}

				_QVL = rtComponent.trafficFlow;
			}
			else if (rtComponent.vehicleType == VehicleType_PL)
			{
				if(rtComponent.flowType != FlowType_START && rtComponent.flowType != FlowType_STOP)
				{
					_Lw_traction_PL = Lw_traction_PL(rtComponent.trafficSpeed, rtComponent.flowType, roadTraffic->ramp);
				}
				else
				{
					double lw = Lw_startStop(rtComponent.vehicleType, roadTraffic->ramp, rtComponent.flowType, roadTraffic->surfaceAge);
					FillRollingAndTractionFromStartStop(lw, rtComponent.vehicleType);
				}

				_QPL = rtComponent.trafficFlow;
			}
		}

		// sum on vehicles :
		double Lw = LwTotal(_Lw_traction_VL, _QVL, _Lw_traction_PL, _QPL);

		// by frequency bands :
		_Lw_band = roadTraffic->_Lw_values ;
		return SpectralDistribution(Lw, roadTraffic->surfaceType);
	}

	/*!
	* \brief Calculation of the traction component of the Sound power level per source line meter for a vehicle type
	*
	* \param vtype The vehicle type
	*
	* \param vehicleSpeed The vehicle speed
	*
	* \param flowType The flow type
	*
	* \param ramp The ramp
	*
	* \returns the sound power associated with the traction component for one single vehicle as a function of the
	*          vehicle type, the vehicle speed, the flow type and the ramp of the road
	*/
	double* CalculRoadEmission::Lw_traction (RoadVehicleType vtype,
											 double vehicleSpeed,
											 RoadFlowType flowType,
											 double ramp)
	{
		double Lm_w_veh = 0;
		switch(vtype)
		{
		case VehicleType_VL :
			Lm_w_veh = Lw_traction_VL(vehicleSpeed, flowType);
			_Lw_traction_VL = Lm_w_veh;
			break;
		case VehicleType_PL:
			Lm_w_veh = Lw_traction_PL(vehicleSpeed, flowType, ramp);
			_Lw_traction_PL = Lm_w_veh;
			break;
		default:
			Lm_w_veh = 0;
			printf("Erreur lors du calcul de la composante moteur : type de vehicule non reconnu \n");
			break;
		}

		// Annexe 4, p.113
		// Correction DVM 09/10/2013: inversion des signes, les fonctions lmentaires du document SETRA
		// permettent de calculer un niveau de pussance par mtre (Lw/m) pour 1 vh/h... Cette fonction
		// retourne la puissance d'un vhicule isol
		Lm_w_veh = Lm_w_veh + 10 * log10(vehicleSpeed) + 30;

		// road surface equals R2 because it may be assimilated with a "reference" road surface
		_Lw_band = _output_values ;
		return SpectralDistribution(Lm_w_veh, RoadSurface_R2);
	}

	/*!
	* \brief Calculates traction component for VL
	*
	* \param vehicleSpeed The vehicle speed
	*
	* \param flowType The flow type
	*
	* \returns the calculated traction component
	*/
	double CalculRoadEmission::Lw_traction_VL (double vehicleSpeed,
											   RoadFlowType flowType)
	{
		if(vehicleSpeed < 20 || vehicleSpeed > 130)
		{
			printf("Erreur lors du calcul de la composante moteur : la vitesse %.2f, utilisee pour un vehicule leger, n'est pas comprise entre 20 et 130 km/h \n", vehicleSpeed);
			return 0;
		}

		double Lm_w_veh = 0;

		//  2.7.3 p. 23-25
		switch(flowType)
		{
		case FlowType_CONST:
			//  2.7.3.1, array (2.9) p. 23
			if (vehicleSpeed <= 30)
			{
				Lm_w_veh = 36.7 - 10. * log10(vehicleSpeed / 90.);
			}
			else if(vehicleSpeed <= 110)
			{
				Lm_w_veh = 42.4 + 2. * log10(vehicleSpeed / 90.);
			}
			else
			{
				Lm_w_veh = 40.7 + 21.3 * log10(vehicleSpeed / 90.);
			}
			break;

		case FlowType_ACC:
			//  2.7.3.2, array (2.11) p. 24
			if (vehicleSpeed < 25)
			{
				printf("Erreur lors du calcul de la composante moteur : la vitesse %.2f, utilisee en acceleration, est inferieure a 25 km/h \n", vehicleSpeed);
				return 0;
			}
			else if (vehicleSpeed <= 100)
			{
				Lm_w_veh = 46.1 - 10. * log10(vehicleSpeed / 90.);
			}
			else
			{
				Lm_w_veh = 44.3 + 28.6 * log10(vehicleSpeed / 90.);
			}
			break;

		case FlowType_DEC:
			//  2.7.3.3, array (2.13) p. 25
			if (vehicleSpeed < 25)
			{
				printf("Erreur lors du calcul de la composante moteur : la vitesse %.2f, utilisee en deceleration, est inferieure a 25 km/h \n", vehicleSpeed);
				return 0;
			}
			else if (vehicleSpeed <= 80)
			{
				Lm_w_veh = 42.1 - 4.5 * log10(vehicleSpeed / 90.);
			}
			else if(vehicleSpeed <= 110)
			{
				Lm_w_veh = 42.4 + 2. * log10(vehicleSpeed / 90.);
			}
			else
			{
				Lm_w_veh = 40.7 + 21.3 * log10(vehicleSpeed / 90.);
			}
			break;

		default:
			Lm_w_veh = 0;
			break;
		}

		return Lm_w_veh;
	}

	/*!
	 * \brief Calculates traction component for PL
	 *
	 * \param vehicleSpeed The vehicle speed
	 *
	 * \param flowType The flow type
	 *
	 * \param ramp The ramp
	 *
	 * \returns the calculated traction component
	 */
	double CalculRoadEmission::Lw_traction_PL (double vehicleSpeed,
											   RoadFlowType flowType, double ramp)
	{
		if(vehicleSpeed < 20 || vehicleSpeed > 100)
		{
			printf("Erreur lors du calcul de la composante moteur : la vitesse %.2f, utilisee pour un poids lourd, n'est pas comprise entre 20 et 100 km/h \n", vehicleSpeed);
			return 0;
		}

		//  2.7.3.4, array (2.15) p. 26
		double Lm_w_veh = 0;
		if (vehicleSpeed <= 70)
		{
			Lm_w_veh = 49.6 - 10. * log10(vehicleSpeed / 80.);
		}
		else
		{
			Lm_w_veh = 50.4 + 3. * log10(vehicleSpeed / 80.);
		}

		//  2.7.3.4, array (2.16) p. 26
		double deltaLm = 0;
		switch(flowType)
		{
		case FlowType_CONST:
			if (ramp >= -2 && ramp <= 2)
			{
				deltaLm = 0.;
			}
			else if(ramp >= 2 && ramp <= 6)
			{
				deltaLm = 2. * (ramp - 2.);
			}
			else if(ramp >= -6 && ramp <= -2)
			{
				deltaLm = (-1.0 * ramp) - 2.; // (-1 * ramp) because ramp < 0
			}
			break;

		case FlowType_ACC:
			deltaLm = 5.;
			if(ramp > 4.5 && ramp <= 6)
			{
				deltaLm += 2. * (ramp - 4.5);
			}
			break;

		case FlowType_DEC:
			deltaLm = 0.;
			if(ramp >= -6 && ramp <= -2)
			{
				deltaLm = (-1.0 * ramp) - 2.; // (-1 * ramp) because ramp < 0
			}
			break;

		default:
			deltaLm = 0;
			break;
		}

		Lm_w_veh += deltaLm;
		return Lm_w_veh;
	}

}
