/*!
* \file CalculRoadEmission.h
* \brief Calculations of road emissions
* \author CSTB
* \version 1.0
*/

#ifndef CALCUL_ROADEMISSION_H
#define CALCUL_ROADEMISSION_H

#include "RoadEmissionNMPB08.h"
#include "RoadSurfaceDescription.h"
#include <vector>
#include <math.h>
#include <map>
using namespace std;
using namespace RoadSurfaceDescriptionNMPB;

namespace CalculRoadEmissionNMPB
{
	/*!
	* \brief Definition of a new road surface type with data needed for rolling component and for spectrum
	* 
	* \param idRoadSurface The road surface id
	* 
	* \param AVL The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	* 
	* \param BVL The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	* 
	* \param Vref_VL The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	* 
	* \param APL The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	* 
	* \param BPL The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	* 
	* \param Vref_PL The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	* 
	* \param spectrum The spectral distribution for this road surface (if null, non draining default values will be used)
	* 
	* \returns true if all OK
	*/
	bool DefineRoadSurfaceType(int idRoadSurface, 
		double AVL, double BVL, double Vref_VL, 
		double APL, double BPL, double Vref_PL,
		double const* spectrum);

	/*!
	 * \brief Class used to road emission calculations
	 */
	class CalculRoadEmission
	{
	public:
		/*!
		* \brief Initialization of CalculRoadEmission class
		*/
		CalculRoadEmission(void);

		/*!
		* \brief Initialization of CalculRoadEmission class with given SpectrumType and the given surface type
		* 
		* \param specType The spectrum type
		* 
		* \param surfType The road surface type
		*/
		CalculRoadEmission(RoadSpectrumType specType, RoadSurfaceType surfType);

		/*!
		* \brief Calculation of the Sound power level per source line meter
		* 
		* \param roadTraffic The road traffic data
		* 
		* \returns the Sound power level per source line meter, for each third-octave band
		*/
		double* SoundPowerLevelPerMeter (RoadTraffic* roadTraffic) ;

		/*!
		* \brief Calculation of the rolling component of the Sound power level per source line meter
		*
		* \param roadTraffic The road traffic data
		* 
		* \returns The rolling component of the Sound power level per source line meter, for each third-octave band
		*/
		double* Lwm_rolling (RoadTraffic* roadTraffic) ;

		/*!
		* \brief Calculation of the traction component of the Sound power level per source line meter
		*
		* \param roadTraffic The road traffic data
		* 
		* \returns the traction component of the Sound power level per source line meter, for each third-octave band
		*/
		double* Lwm_traction (RoadTraffic* roadTraffic) ;

		/*!
		* \brief Calculation of the rolling component of the Sound power level per source line meter for a vehicle type
		* 
		* \param vtype The vehicle type
		* 
		* \param vehicleSpeed The vehicle speed
		* 
		* \param surfaceType The surface type
		* 
		* \param surfaceAge The surface age
		* 
		* \returns the rolling component of the Sound power level per source line meter, for each third-octave band
		*/
		double* Lw_rolling (RoadVehicleType vtype, 
						    double vehicleSpeed, 
							RoadSurfaceType surfaceType, 
							double surfaceAge) ;

		/*!
		* \brief Calculation of the traction component of the Sound power level per source line meter for a vehicle type
		* 
		* \param vtype The vehicle type
		* 
		* \param vehicleSpeed The vehicle speed
		* 
		* \param flowType The flow type
		* 
		* \param ramp The ramp
		* 
		* \returns the traction component of the Sound power level per source line meter, for each third-octave band
		*/
		double* Lw_traction (RoadVehicleType vtype,
							 double vehicleSpeed,
							 RoadFlowType flowType,
							 double ramp) ;

	private:
		/*!
		 * \brief Sound power level per unit of length for a flow rate of one light vehicle/hour (dB/m).
		 */
		double _Lw_VL;
		/*!
		* \brief Sound power level per unit of length for a flow rate of one HGV/hour (dB/m).
		*/
		double _Lw_PL; 
		/*!
		* \brief Rolling component of _Lw_VL.
		*/
		double _Lw_rolling_VL;
		/*!
		* \brief Rolling component of _Lw_PL.
		*/
		double _Lw_rolling_PL;
		/*!
		* \brief Traction component of _Lw_VL.
		*/
		double _Lw_traction_VL;
		/*!
		* \brief Traction component of _Lw_PL.
		*/
		double _Lw_traction_PL;
		/*!
		* \brief Hourly light vehicle flow rate, in vehicles/hour.
		*/
		double _QVL;
		/*!
		* \brief Hourly HGV flow rate, in vehicles/hour.
		*/
		double _QPL;
		/*!
		* \brief Number of frequency bands (18 if third-octave band).
		*/
		int _bandsNumber; 
		/*!
		* \brief True if A spectrum.
		*/
		bool _specA; 
		/*!
		* \brief Sound power level per frequency band.
		*/
		double* _Lw_band;
		/*!
		* \brief Frequency bands (octave, third-octave, ...).
		*/
		const int* _usedFrequencies;
		
		/*!
		 * \brief Linear correction by frequency (for A filter corrections)
		 */
		map<int, double> _linearCorrection;

		/*!
	 	 * \brief local buffer used for output values
		 */
		double	_output_values[18] ;

		/*!
		* \brief : Fill the linear corrections map
		*
		*/
		void FillLinearCorrectionMap();

		/*!
		* \brief If RoadSurfaceType < RoadSurface_UserDefined, add it to the map (if not already in)
		* 
		* \param surfType The surface type id
		*/
		void FillRoadSurfaceDescriptionMap(RoadSurfaceType surfType);

		/*!
		* \brief Give spectral distribution of the power sound value
		* 
		* \param powerValue The sound power value
		* 
		* \param surfaceType The road surface type
		* 
		* \returns an array containing the sound value for each frequency band
		*/
		double* SpectralDistribution(double powerValue, RoadSurfaceType surfaceType);

		/*!
		* \brief Calculation of the Sound power level per source line meter with VL and PL data
		* 
		*  2.3.5, formula (8), p. 13 (without spectral distribution)
		*
		* \param LwVL The Sound power level for VL
		* 
		* \param QVL : average hourly traffic flow for VL
		* 
		* \param LwPL The Sound power level for PL
		* 
		* \param QPL : average hourly traffic flow for PL
		* 
		* \returns the total sound power level
		*/
		double LwTotal(double LwVL, double QVL, double LwPL, double QPL);

		/*!
		* \brief Rolling component of the Sound power level for VL
		* 
		*  2.7.2, arrays (2.6) and (2.8), p. 19-22
		*
		* \param vehicleSpeed The vehicle speed
		* 
		* \param surfaceType The surface type
		* 
		* \param surfaceAge The surface age
		* 
		* \returns the rolling component of the Sound power level per source line meter, for VL
		*/
		double Lw_rolling_VL (double vehicleSpeed, RoadSurfaceType surfaceType, double surfaceAge);

		/*!
		* \brief Rolling component of the Sound power level for PL
		* 
		*  2.7.2, arrays (2.6) and (2.8), p. 19-22
		* 
		* \param vehicleSpeed The vehicle speed
		* 
		* \param surfaceType The surface type
		* 
		* \param surfaceAge The surface age
		* 
		* \returns the rolling component of the Sound power level per source line meter, for PL
		*/
		double Lw_rolling_PL (double vehicleSpeed, RoadSurfaceType surfaceType, double surfaceAge);

		/*!
		* \brief Calculates traction component for VL
		* 
		* \param vehicleSpeed The vehicle speed
		* 
		* \param flowType The flow type
		* 
		* \returns the calculated traction component
		*/
		double Lw_traction_VL(double vehicleSpeed, RoadFlowType flowType);

		/*!
		* \brief Calculates traction component for PL
		* 
		* \param vehicleSpeed The vehicle speed
		* 
		* \param flowType The flow type
		* 
		* \param ramp The ramp
		* 
		* \returns the calculated traction component
		*/
		double Lw_traction_PL(double vehicleSpeed, RoadFlowType flowType, double ramp);

		/*!
		* \brief Calculation of the Sound power level for Start or Stop section
		*
		*  2.7.4, array (2.22-A) and (2.22-B), p. 29
		* 
		* \param vtype The vehicle type
		* 
		* \param ramp The ramp
		* 
		* \param flowType The flow type
		*
		* \param surfaceAge The surface age
		* 
		* \returns the calculated sound power level for start or stop section
		*/
		double Lw_startStop(RoadVehicleType vtype, double ramp, RoadFlowType flowType, double surfaceAge);
		/*!
		* \brief Fill the 2 components _Lw_rolling_veh and _Lw_traction_veh from the calculated Sound power level for Start or Stop section
		* 
		* \param lwStartStop The calculated Sound power level for Start or Stop section
		* 
		* \param vtype The vehicle type
		*/
		void FillRollingAndTractionFromStartStop(double lwStartStop, RoadVehicleType vtype);

		/*!
		* \brief Energetic addition
		*
		*  2.7.1.1, p. 19
		* 
		* \param val1 The first value to add
		* 
		* \param val2 The second value to add
		* 
		* \returns The energetic addition of val1 and val2
		*/
		double EnergeticAddition(double val1, double val2);
	};
}

#endif