/*!
* \file RoadSurfaceDescription.h
* \brief RoadSurfaceDescription Class used to define a road surface
* \author CSTB
* \version 1.0
*/

#ifndef ROADSURFACE_DESCRIPTION_H
#define ROADSURFACE_DESCRIPTION_H

#include <vector>
#include <map>
using namespace std;

namespace RoadSurfaceDescriptionNMPB
{
	/*!
	 * \brief Class used to define a road surface
	 */
	class RoadSurfaceDescription
	{
		public:
			/*!
			* \brief Initialization of RoadSurfaceDescription class
			*/
			RoadSurfaceDescription(void);

			/*!
			* \brief Initialization of RoadSurfaceDescription class with given data  
			* 
			* \param AVL The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			* 
			* \param BVL The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			* 
			* \param Vref_VL The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			* 
			* \param APL The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			* 
			* \param BPL The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			* 
			* \param Vref_PL The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			* 
			* \param spectrum The spectral distribution for this road surface 
			* 
			* \param doNormalization If true, do normalization
			*/
			RoadSurfaceDescription(double AVL, double BVL, double Vref_VL, double APL, double BPL, double Vref_PL, const double* spectrum, bool doNormalization);

			/*!
			* \brief Gets the AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			*/
			double get_AVL()
			{
				return _AVL;
			}
			/*!
			* \brief Gets the BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			*/
			double get_BVL()
			{
				return _BVL;
			}
			/*!
			* \brief Gets the Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			*/
			double get_Vref_VL()
			{
				return _Vref_VL;
			}
			/*!
			* \brief Gets the APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			*/
			double get_APL()
			{
				return _APL;
			}
			/*!
			* \brief Gets the BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			*/
			double get_BPL()
			{
				return _BPL;
			}
			/*!
			* \brief Gets the Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			*/
			double get_Vref_PL()
			{
				return _Vref_PL;
			}
			/*!
			* \brief Gets the Sound power level by frequency
			*/
			map<int,double> get_spectralDistributionMap()
			{
				return _spectralDistributionMap;
			}

		private:
			/*!
			 * \brief The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			 */
			double _AVL;
			/*!
			* \brief The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			*/
			double _BVL;
			/*!
			* \brief The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
			*/
			double _Vref_VL;
			/*!
			* \brief The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			*/
			double _APL;
			/*!
			* \brief The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			*/
			double _BPL;
			/*!
			* \brief The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
			*/
			double _Vref_PL;

			/*!
			* \brief Sound power level by frequency
			*/
			map<int,double> _spectralDistributionMap;

			/*!
			* \brief Spectrum normalization to 0
			*/
			void SpectrumNormalization();
	};

}
#endif